#include "conntrack_kmodule.h"

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Ramana" );
MODULE_DESCRIPTION("WMN: Flow Counter");

int timeInval = 10;
module_param(timeInval, int, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);
MODULE_PARM_DESC(timeInval, "Update Timer Duration");

int maxTtl = 10;
module_param(maxTtl, int, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);
MODULE_PARM_DESC(maxTtl, "Time To Live for a Flow");

int ageToQualify = 0;
module_param(ageToQualify, int, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP);
MODULE_PARM_DESC(ageToQualify, "Minimum Age to Qualify as a Long Lived Flow");

nf_hookfn post_route_handler;
static __u32 numConn;

struct timer_list timer;
spinlock_t wmn_post_route_lock=SPIN_LOCK_UNLOCKED;
 struct nf_hook_ops post_route = {
  {NULL, NULL},
  post_route_handler,
  THIS_MODULE,
  PF_INET,
  NF_IP_POST_ROUTING,
  NF_IP_PRI_CONNTRACK_DEFRAG-10 
};

void connTTLUpdateNow(unsigned long data) {
  	int ind=0;
        struct timeval temp = {timeInval,0};
	struct timeval now;
	unsigned int numFlows=0;
  	for(;ind<numConn;ind++) {
    		if(connList[ind].time2live > timeInval) {
			do_gettimeofday(&now);
		      connList[ind].time2live = connList[ind].time2live - timeInval;
		      connList[ind].connAge = connList[ind].connAge + timeInval;
	 	    printk(KERN_INFO"\n conn#: %d\tsrc: %d.%d.%d.%d:%u\tdst: %d.%d.%d.%d:%u\tttl: %d\tage: %d\ttimenow: %ld:%ld",numConn,NIPQUAD(connList[ind].saddr), ntohs(connList[ind].sport), NIPQUAD(connList[ind].daddr), ntohs(connList[ind].dport), connList[ind].time2live, connList[ind].connAge, now.tv_sec,now.tv_usec); 
		     if(connList[ind].connAge > ageToQualify)
				numFlows++;
		}
		else {
		// delete an entry now
		int i;
		  for(i=ind; i < numConn-1; i++) {
		   connList[i] = connList[i+1];
		  }
		  connList[ind+1].time2live = connList[ind+1].time2live - timeInval;
		  connList[ind+1].connAge = connList[ind+1].connAge + timeInval;
		     if(connList[ind+1].connAge > ageToQualify)
				numFlows++;
		  numConn--;
  	        }
	}
	printk(KERN_CRIT"\nconn#: %d\t",numFlows);
        mod_timer((&timer), jiffies + timeval_to_jiffies(&temp));
}

int search_connection(struct sk_buff **skb) {
  int ind=0;
  int found=-1;
  struct iphdr* iph=(*skb)->nh.iph;
  struct tcphdr *tcph = NULL; 
  struct udphdr *udph = NULL; 
  __be16 source=0;
  __be16 dest=0;
   if(iph->protocol == TCP) {
	tcph = (struct tcphdr *)((*skb)->data + (iph->ihl << 2));
	source = tcph->source;
	dest = tcph->dest;
    } else if(iph->protocol == UDP){
	udph = (struct udphdr *)((*skb)->data + (iph->ihl << 2));
	source = udph->source;
	dest = udph->dest;
   }
  for(;ind<numConn;ind++) {
   if(connList[ind].saddr == iph->saddr && connList[ind].daddr == iph->daddr && connList[ind].sport == source && connList[ind].dport == dest)
     found=ind;
  }
  return found;
}


int insert_data(struct sk_buff **skb) {

	int retVal1 = -1;
	int retVal2 = -1;
	struct iphdr* iph=(*skb)->nh.iph;
	struct tcphdr *tcph = NULL; 
	struct udphdr *udph = NULL; 
	__be16 source=0;
	__be16 dest=0;
	int ind;
	if ((iph->protocol != TCP && iph->protocol != UDP) || (*skb)->len < 100)
		return -1;
	if(iph->protocol == TCP) {
		tcph = (struct tcphdr *)((*skb)->data + (iph->ihl << 2));
		source = tcph->source;
		dest = tcph->dest;
	}
	else if(iph->protocol == UDP){
		udph = (struct udphdr *)((*skb)->data + (iph->ihl << 2));
		source = udph->source;
		dest = udph->dest;
	}
	
	retVal1 = search_connection(skb);
 	retVal2 = search_connection(skb); 

	if(retVal1 >= 0) {
		ind=retVal1;
      		connList[retVal1].time2live=maxTtl; // refresh life time
	}
 	else if(retVal2 >= 0) {
 		ind=retVal2;
       		connList[retVal2].time2live=maxTtl;
	} else {
		if(numConn < MAXCONN) {
	   	  connList[numConn].saddr = iph->saddr;
	   	  connList[numConn].daddr = iph->daddr;
	   	  connList[numConn].sport = source;
	   	  connList[numConn].dport = dest;
      		  connList[numConn].time2live=maxTtl;
      	 	  connList[numConn].connAge=0;
		  ind=numConn;
		  numConn++;
		} else {
		  printk(KERN_INFO"exceeding storage space");
		  return -2;
		}
	}
	return ind;
}

unsigned int post_route_handler(unsigned int hooknum,
                              struct sk_buff **skb,
                              const struct net_device *in,
                              const struct net_device *out,
                              int (*okfn)(struct sk_buff *)){
	int verdict;
	struct iphdr *iph;
	struct net_device * netdev;
	struct in_device * indev;
       __be32 ethIP;

	spin_lock_bh(&wmn_post_route_lock);
	iph=(*skb)->nh.iph;
	netdev = dev_get_by_name("eth0");
	indev = in_dev_get(netdev);
        ethIP = indev->ifa_list->ifa_local;

	if(iph->saddr != ethIP)
	        insert_data(skb);	

	verdict = NF_ACCEPT;
	spin_unlock_bh(&wmn_post_route_lock);
	return verdict;
}	
/* initialise the module */
int wmn_init(void)
{
  struct timeval temp = {timeInval,0};
  int size=sizeof(struct connection)*MAXCONN;
  connList=(struct connection* ) kmalloc(size, GFP_ATOMIC);
  init_timer(&timer);
  timer.function = connTTLUpdateNow;
  timer.expires = jiffies + timeval_to_jiffies(&temp);
  timer.data = (unsigned long)NULL;
  add_timer(&timer);
  nf_register_hook(&post_route);
  spin_lock_init(&wmn_post_route_lock);
  printk(KERN_CRIT"\nConnection Tracking Module Loaded Successfully");
  return 0;
}

/* The legitimate exit routine for the module*/
void wmn_exit(void) {
	nf_unregister_hook(&post_route);
	del_timer(&timer);
  	printk(KERN_CRIT"\nConnection Tracking Module Unloaded Successfully");
}


module_init(wmn_init);
module_exit(wmn_exit);
